import os

import psutil

import ltrim.profiler.profiler as profiler
from ltrim.utils import MB


def get_memory_usage():
    """
    Get the memory usage of the current process in MB.
    """
    process = psutil.Process(os.getpid())
    return process.memory_info().rss / MB


__all__ = ["profiler", "get_memory_usage"]



import os
import sys
import time
from contextlib import contextmanager
from importlib.abc import Loader
from importlib.machinery import PathFinder
from types import MethodType, ModuleType
from typing import Sequence

import psutil

from ltrim.utils.constants import MB, MS


@contextmanager
def profiler_context(profiler, operation_name, counter):
    """
    Context manager for profiling memory and execution time.
    """
    process = psutil.Process(os.getpid())
    start_time = time.time()
    memory_before = process.memory_info().rss

    yield  # This allows the context to be used around any code block

    memory_after = process.memory_info().rss
    end_time = time.time()

    memory_footprint = memory_after - memory_before
    time_taken = end_time - start_time

    profiler[operation_name] = {
        "order": counter,
        "memory": memory_footprint / MB,
        "time": time_taken / MS,
    }


def create_profiler_loader(
    loader: Loader, profiler: dict[str, dict[str, float]], counter: int
):
    """
    Create a loader with profiler attached, based on the original loader
    """

    loader_exec_module = loader.exec_module

    def profiler_exec_module(self, module: ModuleType):
        """
        Execute the module and profile the memory, time footprint
        """
        module_name = module.__name__

        # only record one time
        if module_name not in profiler:
            with profiler_context(profiler, module_name, counter):
                loader_exec_module(module)
        else:
            loader_exec_module(module)

    loader.exec_module = MethodType(profiler_exec_module, loader)

    return loader


class ProfilerMetaFinder(PathFinder):
    """
    Custom meta path finder to time the import of modules
    """

    profiler_report: dict[str, dict[str, float]] = {}
    counter = 0

    @classmethod
    def find_spec(
        cls,
        fullname: str,
        path: Sequence[str] | None = None,
        target: ModuleType | None = None,
    ):
        """
        Find the spec for the given module, delegate to the
        original implementation and replace with our own loader
        """

        spec = super().find_spec(fullname, path, target)
        if spec and spec.loader:
            cls.counter += 1
            spec.loader = create_profiler_loader(
                spec.loader, cls.profiler_report, cls.counter
            )
        return spec


def attach():
    sys.meta_path.insert(0, ProfilerMetaFinder())


def detach():
    sys.meta_path = [i for i in sys.meta_path if not isinstance(i, ProfilerMetaFinder)]


def get_report():
    """
    Return the profiler report
    """
    return ProfilerMetaFinder.profiler_report



from ltrim.transformers.ast_transformers import (
    ImportsFinder,
    RemoveAttribute,
    SetFix,
)
from ltrim.transformers.utils import add_tag, retrieve_name

USED = True
NOT_USED = False

__all__ = [
    # utilities
    "retrieve_name",
    "add_tag",
    # AST transformers
    "ImportsFinder",
    "RemoveAttribute",
    "SetFix",
    # constants
    "USED",
    "NOT_USED",
]



import ast
import logging
import sys
from pprint import pformat as pp

from ltrim.debloat.process import debloat, run_profiler, run_pycg
from ltrim.debloat.utils import (
    blacklist,
    filter_pycg,
    sort_report,
    update_alive_modules,
)
from ltrim.transformers import ImportsFinder
from ltrim.utils import Stats, cmd_message

logger = logging.getLogger(__name__)


class Debloater:
    """
    Debloater class that encapsulates the debloating process.

    :param filename: The name of the application to debloat
    :param top_k: The number of modules to debloat
    :param scoring: The scoring method to calculate the top K ranking of the modules
    """

    def __init__(self, filename, top_K, scoring, disable_pycg, testcases="data.json"):
        self.appname = filename
        self.top_K = top_K
        self.scoring = scoring
        self.stats = Stats(self.appname, self.top_K)
        self.pycg = not disable_pycg
        self.testcases = testcases

    def run(self):
        with open(self.appname, "r") as f:
            source = f.read()

        ast_source = ast.parse(source)

        # --------------------------------------------------------------------- #
        # ----------------------- Static Analysis Phase ----------------------- #
        # --------------------------------------------------------------------- #

        # Step 1: Find imported modules
        cmd_message("Finding imports...")
        imports_finder = ImportsFinder()
        imports_finder.visit(ast_source)
        cmd_message("Imports found!", "success")
        logger.info(f"Imports found: {imports_finder.imports}")

        # Step 2: Use PyCG to extract the call graph of the application
        if sys.version_info.minor <= 10 and self.pycg:
            cmd_message("Extracting call graph...")
            call_graph = run_pycg(self.appname)
            cmd_message("Call graph extracted!", "success")
            logger.info(f"Call graph extracted: {call_graph}")
        else:
            call_graph = []

        # --------------------------------------------------------------------- #
        # ------------------------ Constructing Phase ------------------------- #
        # --------------------------------------------------------------------- #

        # Step 3: Construct a list of the imported modules
        imported_modules = []

        for module_name in imports_finder.imports:
            if module_name in blacklist:
                continue

            imported_modules.append(module_name)

        logger.info(f"Filtered imports: {imported_modules}")

        # --------------------------------------------------------------------- #
        # ------------------------- Profiling Phase --------------------------- #
        # --------------------------------------------------------------------- #

        # Step 4: Profile the import process of the application
        cmd_message("Profiling the import process...")
        report = run_profiler(imported_modules)

        # Extract the total memory used by the imported modules
        total_memory = report["total_memory"]
        del report["total_memory"]

        # Extract the total time taken to import modules
        total_time = sum(
            [report[module]["time"] for module in imported_modules if module in report]
        )

        cmd_message(f"Total memory used in the import process: {total_memory:.2f}MB")
        cmd_message(f"Total time taken for the import process: {total_time:.2f}ms")

        # Step 5 - Sort the profiler report using the scoring method
        sorted_report = sort_report(report, self.scoring, total_time, total_memory)

        # Filter modules in the blacklist
        sorted_report = [
            module for module in sorted_report if module[0] not in blacklist
        ]
        logger.info(pp(sorted_report[: self.top_K]))
        cmd_message("Profiling completed!", "success")

        # --------------------------------------------------------------------- #
        # ------------------------- Debloating Phase -------------------------- #
        # --------------------------------------------------------------------- #

        # Step 6 - Debloat the top K modules

        modules_to_debloat = []

        # Initialize a ModuleRecord for stats tracking
        for module, entry in sorted_report[: self.top_K]:
            modules_to_debloat.append(module)
            self.stats.add_module(module)
            cmd_message(f"Module {module}: {entry}")
            self.stats.set_profiling_stats(
                module=module, memory=entry["memory"], time=entry["time"], before=False
            )

        alive_modules = set(modules_to_debloat)

        for midx, module in enumerate(modules_to_debloat):
            cmd_message(f"Debloating module {module} ({midx + 1}/{self.top_K})", "info")

            if module not in alive_modules:
                cmd_message(f"Module {module} is not longer needed!")
                continue

            # Step 6.1 - Filter the PyCG attributes
            filtered_attributes = filter_pycg(module, call_graph)
            logger.info(f"Attributes to keep based on PyCG: {filtered_attributes}")

            # Step 6.2 - Debloat the module
            module_path, delta_record = debloat(
                self.appname, module, filtered_attributes
            )
            self.stats.set_path(module, module_path)
            self.stats.set_debloating_stats(module, delta_record)

            # Re-import to update alive_modules
            new_report = run_profiler(imported_modules)
            update_alive_modules(alive_modules, new_report)

        # --------------------------------------------------------------------- #
        # ---------------------- Stats-collecting Phase ----------------------- #
        # --------------------------------------------------------------------- #

        # Step 7 - Collect statistics

        # Step 7.1 - Run profiler
        final_report = run_profiler(imported_modules)
        logger.info("Final report after debloating:")
        logger.info(pp(final_report))

        # Step 7.2 - Get statistics for alive modules
        for module in alive_modules:
            entry = final_report[module]
            self.stats.set_profiling_stats(
                module=module, memory=entry["memory"], time=entry["time"], before=True
            )

        # Step 7.3 - Convert to CSV
        self.stats.convert_to_csv()

        # --------------------------------------------------------------------- #
        # ------------------------------- Done -------------------------------- #
        # --------------------------------------------------------------------- #

        cmd_message("Debloating process completed successfully!", "success")

        cmd_message("For detailed statistics, check log/ directory")



"""
Modify modules by removing attributes
"""

import ast
import copy
import importlib
import inspect
import os
import sys

from ltrim.transformers import RemoveAttribute, SetFix
from ltrim.utils import MAGIC_ATTRIBUTES, cp

DEBUG = False


def tag_members(members):
    """
    Tag attributes as class, function, module or variable
    e.g. (m, <module 'm' from '/path/to/module/m'>) -> (m, 'module')
    """

    def tag(t):
        """
        Tag entries

        :param t: Tuple of (member, value)
        """
        member, val = t
        if inspect.isclass(val):
            return (member, "class")
        if inspect.isfunction(val):
            return (member, "function")
        if inspect.ismodule(val):
            return (member, "module")
        return (member, "variable")

    return list(map(tag, members))


class Moduify:
    """
    Instance of modifier

    :param module_name: Name of the module to modify
    :param marked_attributes: Attributes that must be kept
    """

    def __init__(self, module_name, marked_attributes):
        self.module_name = module_name

        # Load the module
        module = importlib.import_module(module_name)

        # We only need the members that are also in dir(module)
        self.members = module.__dict__

        # Get the path of the module
        try:
            self.module_path = self.members["__file__"]
        except KeyError:
            self.module_path = inspect.getfile(module)

        # Compute set of attributes that cannot be removed
        self.needed_attributes = set(marked_attributes + MAGIC_ATTRIBUTES)

        # Get the complete set of attributes
        self.members = {
            key: value for key, value in self.members.items() if key in dir(module)
        }

        # Get the basename of the module
        self.basename = os.path.basename(self.module_path)

        # Retrieve the AST of the module or None if it isn't a Python file
        if os.path.splitext(self.module_path)[1] == ".py":
            self.ast = ast.parse(inspect.getsource(module))
        else:
            self.ast = None

        # Create a backup directory and copy the module to it
        self.backup_dir = os.path.abspath("tmp/" + module_name)
        cp(self.module_path, self.backup_dir + "/" + self.basename)

    def is_python_module(self):
        """
        Check if the module is a Python file by checking if self.ast is
        set (not None)
        """

        return self.ast is not None

    def modify(self, attributes: list, remove=False):
        """
        Modify the module by removing the attributes

        :param attributes: List of attributes to modify
        :param remove: If True, remove the attributes, otherwise keep them
        """

        # Compute the members that need to be removed

        # Filter out the attributes that are needed
        # (magic attributes + marked attributes)
        filtered_members = [
            (member, value)
            for member, value in self.members.items()
            if member not in self.needed_attributes
        ]

        if remove:
            members_to_remove = [
                (member, value)
                for member, value in filtered_members
                if member in attributes
            ]
        else:
            members_to_remove = [
                (member, value)
                for member, value in filtered_members
                if member not in attributes
            ]

        try:
            # Copy the module from the backup directory
            cp(self.backup_dir + "/" + self.basename, self.module_path)
        except Exception as e:
            print(f"Error copying module source: {e}")
            sys.exit(1)

        module_ast = copy.deepcopy(self.ast)

        remove_transformer = RemoveAttribute(tag_members(members_to_remove))
        module_ast = remove_transformer.visit(module_ast)

        # TODO: Just for numpy for now
        if self.module_name == "numpy":
            numpyfix = SetFix(members_to_remove)
            module_ast = numpyfix.visit(module_ast)

        if DEBUG:
            print(ast.dump(module_ast, annotate_fields=True, indent=1))

        # Write back the modified module
        with open(self.module_path, "w", encoding="utf-8") as out:
            new_source = ast.unparse(module_ast)
            out.write(new_source)
            out.flush()

        return module_ast

    def restore_original_directory(self):
        """
        Move the backup init file back to the original directory
        """

        original_file = self.backup_dir + "/__init__.py"

        if self.module_name not in sys.stdlib_module_names:
            cp(original_file, self.module_path + "/__init__.py")



import importlib

from pycgl import formats
from pycgl.pycg import CallGraphGenerator
from pycgl.utils.constants import CALL_GRAPH_OP

from ltrim.debloat.utils import isolate
from ltrim.delta import DeltaDebugger
from ltrim.profiler import get_memory_usage, profiler


@isolate
def run_pycg(target):
    """
    Run PyCG to exctract the call graph of the target program.

    :param target: The target program to run
    :param output_dict: The output dictionary to store the call graph
    """
    cg = CallGraphGenerator([target], None, -1, CALL_GRAPH_OP)
    cg.analyze()

    simple_formatter = formats.Simple(cg)
    return list(simple_formatter.generate().keys())


@isolate
def run_profiler(modules):
    """
    Profile the import process of a list of modules.

    :param modules: A list of modules to profile
    """
    profiler.attach()

    starting_memory = get_memory_usage()

    for module in modules:
        importlib.import_module(module)

    ending_memory = get_memory_usage()

    profiler_report = profiler.get_report()
    profiler_report["total_memory"] = ending_memory - starting_memory

    profiler.detach()

    return profiler_report


@isolate
def debloat(target, module, marked_attributes):
    """
    Run the DeltaDebugger to debloat a module imported from the target program.

    :param target: The target program to run
    :param module: The module to debloat
    :param marked_attributes: The marked attributes to keep
    """

    delta_debugger = DeltaDebugger(target, module, marked_attributes)

    debloated_attributes, delta_record = delta_debugger.delta_debug(log=True)

    module_path = delta_debugger.finalize_module(debloated_attributes)

    return module_path, delta_record



from ltrim.utils._io import cp, mkdirp
from ltrim.utils.constants import MAGIC_ATTRIBUTES, MB, MS
from ltrim.utils.printing import cmd_message
from ltrim.utils.stats import DeltaRecord, ModuleRecord, Stats

__all__ = [
    # Constants
    "MB",
    "MS",
    "MAGIC_ATTRIBUTES",
    # Bash commands
    "cp",
    "mkdirp",
    # Records types and classes
    "DeltaRecord",
    "ModuleRecord",
    "Stats",
    # Printing functions
    "cmd_message",
]



import os
import subprocess

driver_path = os.path.dirname(__file__) + "/driver.py"


class Found(Exception):
    pass


class PyLambdaRunner:
    """
    Class to run a Python file with a handler function and test cases.

    :param file_path: The path to the Python file to run
    :param handler: The handler function to run
    :param test_cases: The test cases to run
    """

    def __init__(self, file_path, handler="handler", test_cases="data.json"):
        self.file_path = file_path
        self.handler = handler
        self.test_cases = test_cases

    def run(self):
        try:
            process = subprocess.run(
                [
                    "python",
                    driver_path,
                    self.file_path,
                    "--handler",
                    self.handler,
                    "--test",
                    self.test_cases,
                ],
                capture_output=True,
                check=True,
            )

            return process
        except subprocess.CalledProcessError as e:
            return e


def chunks(xs, n):
    """
    Yield n chunks from xs.

    :param xs: The list to chunk
    :param n: The number of the chunks
    """
    k, m = divmod(len(xs), n)
    return (
        # fmt: off
        xs[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)]
        for i in range(n)
        # fmt: on
    )


def flatten(lst):
    """
    Flatten a list of lists

    :param lst: The list to flatten
    """
    return [item for sublist in lst for item in sublist]



import ast


def retrieve_name(node):
    """
    Retrieve the name of a node.
    """
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Attribute):
        value_name = retrieve_name(node.value)
        ret = node.attr
        if value_name is not None:
            ret = value_name + "." + node.attr
        return ret
    if isinstance(node, ast.Call):
        return retrieve_name(node.func)
    if isinstance(node, ast.JoinedStr):
        joined = ""
        for val in node.values:
            if isinstance(val, ast.FormattedValue):
                joined += retrieve_name(val.value)
            elif isinstance(val, ast.Constant):
                joined += val.value
        return joined
    if isinstance(node, ast.Constant):
        return node.value
    return None


def add_tag(entries, tag):
    """
    Mark entries with USED or NOT_USED tags.
    """
    return map(lambda entry: (entry[0], (entry[1], tag)), entries)
